# Produces the task-bar icon.  
from win32api import *
from win32gui import *
import win32con
import sys, os, time, traceback
import threading
from twisted.internet import reactor
import Processor

class SysTray(threading.Thread):

    def __init__(self, reader, console, processors):
        self.reader = reader
        self.console = console
        self.processors = processors
        threading.Thread.__init__(self)
        
    def run(self):
        msg_TaskbarRestart = RegisterWindowMessage("TaskbarCreated");
        message_map = {
                msg_TaskbarRestart: self.OnRestart,
                win32con.WM_DESTROY: self.OnDestroy,
                win32con.WM_COMMAND: self.OnCommand,
                win32con.WM_USER+20 : self.OnTaskbarNotify,
        }
        # Register the Window class.
        wc = WNDCLASS()
        hinst = wc.hInstance = GetModuleHandle(None)
        wc.lpszClassName = "cvubot"
        wc.style = win32con.CS_VREDRAW | win32con.CS_HREDRAW;
        wc.hCursor = LoadCursor( 0, win32con.IDC_ARROW )
        wc.hbrBackground = win32con.COLOR_WINDOW
        wc.lpfnWndProc = message_map # could also specify a wndproc.
        classAtom = RegisterClass(wc)
        # Create the Window.
        style = win32con.WS_OVERLAPPED | win32con.WS_SYSMENU
        self.hwnd = CreateWindow( classAtom, "cvubot", style, \
                0, 0, win32con.CW_USEDEFAULT, win32con.CW_USEDEFAULT, \
                0, 0, hinst, None)
        UpdateWindow(self.hwnd)
        self._DoCreateIcons()
        PumpMessages()
        reactor.stop()

    def _DoCreateIcons(self):
        # Try and find a custom icon
        try:
            hinst =  GetModuleHandle(None)
            iconPath = "CVUBot/R2-D2.ico"
            if os.path.isfile(iconPath):
                icon_flags = win32con.LR_LOADFROMFILE | win32con.LR_DEFAULTSIZE
                hicon = LoadImage(hinst, iconPath, win32con.IMAGE_ICON, 0, 0, icon_flags)
            else:
                hicon = LoadIcon(0, win32con.IDI_APPLICATION)

            flags = NIF_ICON | NIF_MESSAGE | NIF_TIP
            nid = (self.hwnd, 0, flags, win32con.WM_USER+20, hicon, "cvubot")
            Shell_NotifyIcon(NIM_ADD, nid)
        except:
            pass

    def OnRestart(self, hwnd, msg, wparam, lparam):
        self._DoCreateIcons()

    def OnDestroy(self, hwnd, msg, wparam, lparam):
        nid = (self.hwnd, 0)
        Shell_NotifyIcon(NIM_DELETE, nid)
        PostQuitMessage(0) # Terminate the app.

    def OnTaskbarNotify(self, hwnd, msg, wparam, lparam):
        if lparam==win32con.WM_RBUTTONUP:
        
            menu = CreatePopupMenu()
            AppendMenu( menu, win32con.MF_STRING, 1025, "Shutdown CVUBot" )
            AppendMenu( menu, win32con.MF_STRING, 1026, "Reset IRC" )
            AppendMenu( menu, win32con.MF_SEPARATOR, 0, "")
            i=0		
            self.maplangs = {}
            for fname in os.listdir("Config/"):
                if fname == "Base" or not os.path.isdir("Config/%s"%fname):
                    continue
                menu2 = CreatePopupMenu()
                if fname in self.processors:
                    AppendMenu(menu2, win32con.MF_STRING, 2000+i, "Stop")
                else:
                    AppendMenu(menu2, win32con.MF_STRING, 2100+i, "Start")
                self.maplangs[2000+i]=fname
                i = i + 1
                AppendMenu( menu, win32con.MF_STRING+win32con.MF_POPUP, menu2, fname)

            pos = GetCursorPos()
            SetForegroundWindow(self.hwnd)
            TrackPopupMenu(menu, win32con.TPM_LEFTALIGN, pos[0], pos[1], 0, self.hwnd, None)
            PostMessage(self.hwnd, win32con.WM_NULL, 0, 0)

        return 1

    def OnCommand(self, hwnd, msg, wparam, lparam):
        id = LOWORD(wparam)
        if id == 1025:
            DestroyWindow(self.hwnd)
        if id == 1026:
            try:
                self.console.lastp.reset()
            except:
                pass
            try:
                self.reader.lastp.reset()
            except:
                pass
        if id >= 2000 and id < 2100:
            try:
                lang = self.maplangs[id]
                proc = self.processors.get(lang)
                self.reader.lastp.removelang(lang, proc)
                proc.unload()
                self.console.lastp.delProcessor(lang, proc)
                self.processors.pop(lang, None)
            except:
                traceback.print_exc()
            return
        if id >= 2100:
            try:
                lang = self.maplangs[id]
                processor = self.factory.processors.get(lang, None)
                if processor != None:
                    return
                processor = Processor.Processor(lang)
                self.processors[lang]=processor
                self.console.lastp.addProcessor(lang)
                self.reader.lastp.addlang(lang)
            except:
                pass
  
  
if __name__=='__main__':
    x = SysTray(None, None, {})
    x.start()
    print "running"

